#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Optimization;
using System;

namespace HeuristicLab.Problems.DataAnalysis {
  /// <summary>
  /// Represents a classification solution that uses a discriminant function and classification thresholds.
  /// </summary>
  [StorableClass]
  [Item("DiscriminantFunctionClassificationSolution", "Represents a classification solution that uses a discriminant function and classification thresholds.")]
  public class DiscriminantFunctionClassificationSolution : ClassificationSolution, IDiscriminantFunctionClassificationSolution {
    [StorableConstructor]
    protected DiscriminantFunctionClassificationSolution(bool deserializing) : base(deserializing) { }
    protected DiscriminantFunctionClassificationSolution(DiscriminantFunctionClassificationSolution original, Cloner cloner)
      : base(original, cloner) {
    }
    public DiscriminantFunctionClassificationSolution(IRegressionModel model, IClassificationProblemData problemData)
      : this(new DiscriminantFunctionClassificationModel(model, problemData.ClassValues), problemData) {
    }
    public DiscriminantFunctionClassificationSolution(IDiscriminantFunctionClassificationModel model, IClassificationProblemData problemData)
      : base(model, problemData) {
      Model.ThresholdsChanged += new EventHandler(Model_ThresholdsChanged);
    }

    #region IDiscriminantFunctionClassificationSolution Members

    public new IDiscriminantFunctionClassificationModel Model {
      get { return (IDiscriminantFunctionClassificationModel)base.Model; }
    }

    public IEnumerable<double> EstimatedValues {
      get { return GetEstimatedValues(Enumerable.Range(0, ProblemData.Dataset.Rows)); }
    }

    public IEnumerable<double> EstimatedTrainingValues {
      get { return GetEstimatedValues(ProblemData.TrainingIndizes); }
    }

    public IEnumerable<double> EstimatedTestValues {
      get { return GetEstimatedValues(ProblemData.TestIndizes); }
    }

    public IEnumerable<double> GetEstimatedValues(IEnumerable<int> rows) {
      return Model.GetEstimatedValues(ProblemData.Dataset, rows);
    }

    public IEnumerable<double> Thresholds {
      get {
        return Model.Thresholds;
      }
      set { Model.Thresholds = new List<double>(value); }
    }

    public event EventHandler ThresholdsChanged;

    private void Model_ThresholdsChanged(object sender, EventArgs e) {
      OnThresholdsChanged(e);
    }

    protected virtual void OnThresholdsChanged(EventArgs e) {
      var listener = ThresholdsChanged;
      if (listener != null) listener(this, e);
    }
    #endregion

    public override IEnumerable<double> GetEstimatedClassValues(IEnumerable<int> rows) {
      if (Model.Thresholds == null || Model.Thresholds.Count() == 0) RecalculateClassIntermediates();
      return base.GetEstimatedClassValues(rows);
    }

    private void RecalculateClassIntermediates() {
      int slices = 100;
      List<double> estimatedValues = EstimatedValues.ToList();
      List<int> classInstances = (from classValue in ProblemData.Dataset.GetVariableValues(ProblemData.TargetVariable)
                                  group classValue by classValue into grouping
                                  select grouping.Count()).ToList();
      double maxEstimatedValue = estimatedValues.Max();
      double minEstimatedValue = estimatedValues.Min();
      List<KeyValuePair<double, double>> estimatedTargetValues =
         (from row in ProblemData.TrainingIndizes
          select new KeyValuePair<double, double>(
            estimatedValues[row],
            ProblemData.Dataset[ProblemData.TargetVariable, row])).ToList();

      List<double> originalClasses = ProblemData.ClassValues.OrderBy(x => x).ToList();
      int nClasses = originalClasses.Distinct().Count();
      double[] thresholds = new double[nClasses + 1];
      thresholds[0] = double.NegativeInfinity;
      thresholds[thresholds.Length - 1] = double.PositiveInfinity;

      double thresholdIncrement = (maxEstimatedValue - minEstimatedValue) / slices;

      for (int i = 1; i < thresholds.Length - 1; i++) {
        double lowerThreshold = thresholds[i - 1];
        double actualThreshold = Math.Max(lowerThreshold, minEstimatedValue);
        double lowestBestThreshold = double.NaN;
        double highestBestThreshold = double.NaN;
        double bestClassificationScore = double.PositiveInfinity;
        bool seriesOfEqualClassificationScores = false;

        while (actualThreshold < maxEstimatedValue) {
          double classificationScore = 0.0;

          foreach (KeyValuePair<double, double> estimatedTarget in estimatedTargetValues) {
            //all positives
            if (estimatedTarget.Value.IsAlmost(originalClasses[i - 1])) {
              if (estimatedTarget.Key > lowerThreshold && estimatedTarget.Key < actualThreshold)
                //true positive
                classificationScore += ProblemData.GetClassificationPenalty(originalClasses[i - 1], originalClasses[i - 1]);
              else
                //false negative
                classificationScore += ProblemData.GetClassificationPenalty(originalClasses[i], originalClasses[i - 1]);
            }
              //all negatives
            else {
              if (estimatedTarget.Key > lowerThreshold && estimatedTarget.Key < actualThreshold)
                //false positive
                classificationScore += ProblemData.GetClassificationPenalty(originalClasses[i - 1], originalClasses[i]);
              else
                //true negative, consider only upper class
                classificationScore += ProblemData.GetClassificationPenalty(originalClasses[i], originalClasses[i]);
            }
          }

          //new best classification score found
          if (classificationScore < bestClassificationScore) {
            bestClassificationScore = classificationScore;
            lowestBestThreshold = actualThreshold;
            highestBestThreshold = actualThreshold;
            seriesOfEqualClassificationScores = true;
          }
            //equal classification scores => if seriesOfEqualClassifcationScores == true update highest threshold
          else if (Math.Abs(classificationScore - bestClassificationScore) < double.Epsilon && seriesOfEqualClassificationScores)
            highestBestThreshold = actualThreshold;
          //worse classificatoin score found reset seriesOfEqualClassifcationScores
          else seriesOfEqualClassificationScores = false;

          actualThreshold += thresholdIncrement;
        }
        //scale lowest thresholds and highest found optimal threshold according to the misclassification matrix
        double falseNegativePenalty = ProblemData.GetClassificationPenalty(originalClasses[i], originalClasses[i - 1]);
        double falsePositivePenalty = ProblemData.GetClassificationPenalty(originalClasses[i - 1], originalClasses[i]);
        thresholds[i] = (lowestBestThreshold * falsePositivePenalty + highestBestThreshold * falseNegativePenalty) / (falseNegativePenalty + falsePositivePenalty);
      }
      Thresholds = new List<double>(thresholds);
    }
  }
}
